/*                
 *                        Licensing Notification
 *
 * Permission to use, reproduce, prepare derivative works, and to redistribute
 * to others this software and its documentation is hereby granted, provided
 * that this notice is retained thereon and on all copies or modifications. UChicago
 * Argonne, LLC, Tech-X Corporation, and all other contributors make no representations
 * as to the suitability and operability of this software for any purpose. It is
 * provided "as is" without express or implied warranty.
 *
 */


#ifndef CSR_TRI_SOLVE_GPU_LEVEL_SCHEDULER_H
#define CSR_TRI_SOLVE_GPU_LEVEL_SCHEDULER_H

// -----------------------------------------------------------------------------

#include "csr_matrix_data.h"
#include "operator_T.h"
#include <cstddef>
#include <stdio.h>
#include <iostream>
#include <algorithm> // std::min
#include <cuda.h>
#ifdef CUSP_USE_TEXTURE_MEMORY
#include <cusp/detail/device/texture.h>
#endif
#include <thrust/detail/device/cuda/arch.h>

// -----------------------------------------------------------------------------
__device__ static unsigned int csr_tri_solve_lev_sched_count;
// -----------------------------------------------------------------------------

// -----------------------------------------------------------------------------

/*
	Kernel function is based on CUSP library's spmv_csr_vector_kernel
	function which is licensed under the Apache 2.0 license.  Here
	is the license information for that part of the code.

 *  Copyright 2008-2010 NVIDIA Corporation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.

*/

// -----------------------------------------------------------------------------
// THE CODE BELOW IS CURRENTLY NOT WORKING ... DO NOT USE

template <typename index_type, typename value_type, unsigned int USECACHE, unsigned int THREADS_PER_VECTOR, unsigned int VECTORS_PER_BLOCK>
__global__ void csr_tri_solve_gpu_level_scheduler_kernel1(
				      const index_type num_rows,
                                      const index_type * Ap, // A.row_offsets
				      const index_type * Aj, // A.column_indices
				      const value_type * Ax, // A.values
				      const index_type nLevels,
				      const index_type maxNumUnknownsAtSameLevel,
				      const index_type levelSum,
				      const index_type* levels,
				      const index_type* perms,
				      volatile value_type* x)
{  

  const index_type THREADS_PER_BLOCK = VECTORS_PER_BLOCK * THREADS_PER_VECTOR;
  const index_type thread_id   = THREADS_PER_BLOCK * blockIdx.x + threadIdx.x;    // global thread index
  const index_type vector_id   = thread_id   /  THREADS_PER_VECTOR;               // global vector index
  //if (vector_id >= maxNumUnknownsAtSameLevel) return;

  const index_type vector_lane = threadIdx.x /  THREADS_PER_VECTOR;               // vector index within the block
  const index_type thread_lane = threadIdx.x & (THREADS_PER_VECTOR - 1);          // thread index within the vector
  const index_type num_vectors = VECTORS_PER_BLOCK * gridDim.x;                   // total number of active vectors

  __shared__ volatile value_type sdata[THREADS_PER_BLOCK + THREADS_PER_VECTOR / 2];  // padded to avoid reduction conditionals
  __shared__ volatile index_type ptrs[VECTORS_PER_BLOCK][2];
  __shared__ volatile index_type permutedIndex[VECTORS_PER_BLOCK];

  index_type start=0;
  __shared__ volatile int numRows;

  if (threadIdx.x==0) csr_tri_solve_lev_sched_count = 0;
  __threadfence();
  
  for(index_type level = 0; level < nLevels; level++) {
    if (threadIdx.x==0)
      numRows = levels[level];
    for(index_type row = vector_id; row < numRows; row += num_vectors) {
      if(thread_lane == 0) permutedIndex[vector_lane] = perms[row+start];
      index_type p = permutedIndex[vector_lane];

      // use two threads to fetch Ap[row] and Ap[row+1]
      // this is considerably faster than the straightforward version
      if(thread_lane < 2)
	ptrs[vector_lane][thread_lane] = Ap[p + thread_lane];

      const index_type row_start = ptrs[vector_lane][0]; //same as: row_start = Ap[row];
      const index_type row_end   = ptrs[vector_lane][1]; //same as: row_end   = Ap[row+1];

      // initialize local sum
      value_type sum = 0.;

      if (THREADS_PER_VECTOR == 32 && row_end - row_start > 32)
	{
	  // ensure aligned memory access to Aj and Ax
	  index_type jj = row_start - (row_start & (THREADS_PER_VECTOR - 1)) + thread_lane;	// accumulate local sums
	  if(jj < row_end)
            sum += Ax[jj] * x[Aj[jj]];
          //sum += Ax[jj] * fetch_x<USECACHE>(Aj[jj], x);

	  // accumulate local sums
	  for(jj += THREADS_PER_VECTOR; jj < row_end; jj += THREADS_PER_VECTOR)
            sum += Ax[jj] * x[Aj[jj]];
          //sum += Ax[jj] * fetch_x<USECACHE>(Aj[jj], x);
	}
      else
	{
	  // accumulate local sums
	  for(index_type jj = row_start+thread_lane; jj < row_end; jj += THREADS_PER_VECTOR) {
            sum += Ax[jj] * x[Aj[jj]];
            //sum += Ax[jj] * fetch_x<USECACHE>(Aj[jj], x);
	  }
	}

      // store local sum in shared memory
      sdata[threadIdx.x] = sum;

      // reduce local sums to row sum
      if (THREADS_PER_VECTOR > 16) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x + 16];
      if (THREADS_PER_VECTOR >  8) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  8];
      if (THREADS_PER_VECTOR >  4) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  4];
      if (THREADS_PER_VECTOR >  2) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  2];
      if (THREADS_PER_VECTOR >  1) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  1];

      // first thread writes the result
      if (thread_lane == 0) {
	x[p] -= sum;
	atomicInc(&csr_tri_solve_lev_sched_count, levelSum);
	__threadfence();
      }
    }

    /* Make sure ALL threads in ALL blocks catch up */
    if (threadIdx.x == 0)
      while(atomicCAS(&csr_tri_solve_lev_sched_count, start+numRows, start+numRows) != start+numRows) ;
    __syncthreads();
    //__threadfence();
    start+=numRows;
  }
}


// THE CODE ABOVE IS CURRENTLY NOT WORKING ... DO NOT USE
// -----------------------------------------------------------------------------
// THE CODE BELOW IS CURRENTLY NOT WORKING ... DO NOT USE

template <unsigned int USECACHE, unsigned int THREADS_PER_VECTOR, typename index_type, typename value_type>
void csr_tri_solve_gpu_level_scheduler_T(const csr_matrix_data<index_type, value_type>& A, 
						const index_type nLevels,
						const index_type maxNumUnknownsAtSameLevel,
						const index_type levelSum,
						const index_type* levels,
						const index_type* perms,
						value_type* x)
{
  if (USECACHE)
    bind_x(x);

  // call the kernel

  const size_t THREADS_PER_BLOCK  = 32;
  const size_t VECTORS_PER_BLOCK  = THREADS_PER_BLOCK / THREADS_PER_VECTOR;

  const size_t MAX_BLOCKS = thrust::detail::device::cuda::arch::max_active_blocks(csr_tri_solve_gpu_level_scheduler_kernel1<index_type, value_type, USECACHE, THREADS_PER_VECTOR, VECTORS_PER_BLOCK>, THREADS_PER_BLOCK, (size_t) 0);
  const size_t NUM_BLOCKS = std::min<size_t>(MAX_BLOCKS, DIVIDE_INTO(maxNumUnknownsAtSameLevel, VECTORS_PER_BLOCK));

  csr_tri_solve_gpu_level_scheduler_kernel1<index_type, value_type, USECACHE, THREADS_PER_VECTOR, VECTORS_PER_BLOCK>
    <<<NUM_BLOCKS, THREADS_PER_BLOCK>>>(A.num_rows,
					A.row_offsets,         // A.row_offsets
					A.column_indices,      // A.column_indices
					A.values,              // A.values
					nLevels,
					maxNumUnknownsAtSameLevel,
					levelSum,
					levels, perms, x);
  if (USECACHE)
    unbind_x(x);
}

// THE CODE ABOVE IS CURRENTLY NOT WORKING ... DO NOT USE



// -----------------------------------------------------------------------------

template <typename index_type, typename value_type, unsigned int USECACHE, unsigned int THREADS_PER_VECTOR, unsigned int VECTORS_PER_BLOCK>
__global__ void csr_tri_solve_gpu_level_scheduler_kernel2(
				      const index_type num_rows,
                                      const index_type * Ap, // A.row_offsets
				      const index_type * Aj, // A.column_indices
				      const value_type * Ax, // A.values
				      const index_type shift,
				      const index_type* perms,
				      value_type* x)
{
  const index_type THREADS_PER_BLOCK = VECTORS_PER_BLOCK * THREADS_PER_VECTOR;
  const index_type thread_id   = THREADS_PER_BLOCK * blockIdx.x + threadIdx.x;    // global thread index
  const index_type vector_id   = thread_id   /  THREADS_PER_VECTOR;               // global vector index

  if (vector_id >= num_rows) return;

  const index_type vector_lane = threadIdx.x /  THREADS_PER_VECTOR;               // vector index within the block
  const index_type thread_lane = threadIdx.x & (THREADS_PER_VECTOR - 1);          // thread index within the vector
  const index_type num_vectors = VECTORS_PER_BLOCK * gridDim.x;                   // total number of active vectors

  __shared__ volatile value_type sdata[VECTORS_PER_BLOCK * THREADS_PER_VECTOR + THREADS_PER_VECTOR / 2];  // padded to avoid reduction conditionals
  __shared__ volatile index_type ptrs[VECTORS_PER_BLOCK][2];
  __shared__ volatile index_type permutedIndex[VECTORS_PER_BLOCK];

  for(index_type row = vector_id; row < num_rows; row += num_vectors) {

    if(thread_lane == 0) permutedIndex[vector_lane] = perms[row+shift];
    index_type p = permutedIndex[vector_lane];

    // use two threads to fetch Ap[row] and Ap[row+1]
    // this is considerably faster than the straightforward version
    if(thread_lane < 2)
      ptrs[vector_lane][thread_lane] = Ap[p + thread_lane];

    const index_type row_start = ptrs[vector_lane][0]; //same as: row_start = Ap[row];
    const index_type row_end   = ptrs[vector_lane][1]; //same as: row_end   = Ap[row+1];

    // initialize local sum
    value_type sum = 0.;

    if (THREADS_PER_VECTOR == 32 && row_end - row_start > 32)
      {
	// ensure aligned memory access to Aj and Ax
	index_type jj = row_start - (row_start & (THREADS_PER_VECTOR - 1)) + thread_lane;	// accumulate local sums
	if(jj < row_end)
          sum += Ax[jj] * fetch_x<USECACHE>(Aj[jj], x);
        //sum += Ax[jj] * x[Aj[jj]];

	// accumulate local sums
	for(jj += THREADS_PER_VECTOR; jj < row_end; jj += THREADS_PER_VECTOR)
          sum += Ax[jj] * fetch_x<USECACHE>(Aj[jj], x);
        //sum += Ax[jj] * x[Aj[jj]];
      }
    else
      {
	// accumulate local sums
	for(index_type jj = row_start+thread_lane; jj < row_end; jj += THREADS_PER_VECTOR) {
          sum += Ax[jj] * fetch_x<USECACHE>(Aj[jj], x);
          //sum += Ax[jj] * x[Aj[jj]];
	}
      }

    // store local sum in shared memory
    sdata[threadIdx.x] = sum;

    // reduce local sums to row sum
    if (THREADS_PER_VECTOR > 16) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x + 16];
    if (THREADS_PER_VECTOR >  8) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  8];
    if (THREADS_PER_VECTOR >  4) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  4];
    if (THREADS_PER_VECTOR >  2) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  2];
    if (THREADS_PER_VECTOR >  1) sdata[threadIdx.x] = sum = sum + sdata[threadIdx.x +  1];

    // first thread writes the result
    if (thread_lane == 0)
      x[p] -= sum;
  }
}


// -----------------------------------------------------------------------------


template <unsigned int USECACHE, unsigned int THREADS_PER_VECTOR, typename index_type, typename value_type>
static void csr_tri_solve_cpu_level_scheduler_T(const csr_matrix_data<index_type, value_type>& A,
						const index_type nLevels,
						const index_type maxNumUnknownsAtSameLevel,
						const index_type levelSum,
						const index_type* levels,
						const index_type* perms,
						value_type* x)
{
  if (USECACHE)
    bind_x(x);

  //cudaDeviceProp deviceProp;
  index_type shift = 0;
  cudaError_t err;
  //int device;

  err = cudaGetLastError();
  if (err!=0) std::cout << cudaGetErrorString(err) << " on entrance!" << std::endl;

  const size_t THREADS_PER_BLOCK  = 32;
  const size_t VECTORS_PER_BLOCK  = THREADS_PER_BLOCK / THREADS_PER_VECTOR;

  const size_t MAX_BLOCKS = 4*thrust::detail::device::cuda::arch::max_active_blocks(csr_tri_solve_gpu_level_scheduler_kernel2<index_type, value_type, USECACHE, THREADS_PER_VECTOR, VECTORS_PER_BLOCK>, THREADS_PER_BLOCK, (size_t) 0);
  const size_t NUM_BLOCKS = std::min<size_t>(MAX_BLOCKS, DIVIDE_INTO(maxNumUnknownsAtSameLevel, VECTORS_PER_BLOCK));
  for (index_type level=0; level<nLevels; level++) {
    // call the kernels
    csr_tri_solve_gpu_level_scheduler_kernel2<index_type, value_type, USECACHE, THREADS_PER_VECTOR, VECTORS_PER_BLOCK>
      <<<NUM_BLOCKS, THREADS_PER_BLOCK>>>(levels[level],
					  A.row_offsets,         // A.row_offsets
					  A.column_indices,      // A.column_indices
					  A.values,              // A.values
					  shift, perms, x);
    shift+=levels[level];
  }
  if (USECACHE)
    unbind_x(x);
}


// -----------------------------------------------------------------------------

template <typename index_type, typename value_type>
  int csr_tri_solve_level_scheduler(const csr_matrix_data<index_type, value_type>* A,
				    const index_type nLevels,
				    const index_type maxNumUnknownsAtSameLevel,
				    const index_type levelSum,
				    const index_type* levels,
				    const index_type* levelsCPU,
				    const index_type* perms,
				    value_type* x)
{
  const index_type nnz_per_row = A->num_entries / A->num_rows;

#ifdef CUSP_USE_TEXTURE_MEMORY
  #define USECACHE 1
#else
  #define USECACHE 0
#endif

#if 1
    // This code launches nLevels kernels on the GPU in stream 0. They are naturally synchronized.
    if (nnz_per_row <=  2)  { csr_tri_solve_cpu_level_scheduler_T<USECACHE,2>(*A, nLevels, maxNumUnknownsAtSameLevel, levelSum, levelsCPU, perms, x); return 0;}
    if (nnz_per_row <=  4)  { csr_tri_solve_cpu_level_scheduler_T<USECACHE,4>(*A, nLevels, maxNumUnknownsAtSameLevel, levelSum, levelsCPU, perms, x); return 0;}
    if (nnz_per_row <=  8)  { csr_tri_solve_cpu_level_scheduler_T<USECACHE,8>(*A, nLevels, maxNumUnknownsAtSameLevel, levelSum, levelsCPU, perms, x); return 0;}
    if (nnz_per_row <=  16) { csr_tri_solve_cpu_level_scheduler_T<USECACHE,16>(*A, nLevels, maxNumUnknownsAtSameLevel, levelSum, levelsCPU, perms, x); return 0;}
    if (nnz_per_row <=  32) { csr_tri_solve_cpu_level_scheduler_T<USECACHE,32>(*A, nLevels, maxNumUnknownsAtSameLevel, levelSum, levelsCPU, perms, x); return 0;}

#else
    // This code launches 1 kernel on the GPU in stream 0. Blocks must be sychronized using atomics. This code is not verified yet.
    if (nnz_per_row <=  2)  { csr_tri_solve_gpu_level_scheduler_T<USECACHE,2>(*A, nLevels, maxNumUnknownsAtSameLevel, levelSum, levels, perms, x); return 0;}
    if (nnz_per_row <=  4)  { csr_tri_solve_gpu_level_scheduler_T<USECACHE,4>(*A, nLevels, maxNumUnknownsAtSameLevel, levelSum, levels, perms, x); return 0;}
    if (nnz_per_row <=  8)  { csr_tri_solve_gpu_level_scheduler_T<USECACHE,8>(*A, nLevels, maxNumUnknownsAtSameLevel, levelSum, levels, perms, x); return 0;}
    if (nnz_per_row <=  16) { csr_tri_solve_gpu_level_scheduler_T<USECACHE,16>(*A, nLevels, maxNumUnknownsAtSameLevel, levelSum, levels, perms, x); return 0;}
    if (nnz_per_row <=  32) { csr_tri_solve_gpu_level_scheduler_T<USECACHE,32>(*A, nLevels, maxNumUnknownsAtSameLevel, levelSum, levels, perms, x); return 0;}

#endif

#if 0
    /**************************************************************************************/
    /* This code is used for diagnostics ... i.e. for comparing the results of the single */
    /* versus multiple kernel approach to doing the level scheduling algorithm. It is not */
    /* used at the moment.                                                                */
    /**************************************************************************************/

    std::cout << "nnz_per_row="<<nnz_per_row<<std::endl;
    value_type * xtemp;
    cudaMalloc((void **) &xtemp, A->num_rows*sizeof(value_type));
    cudaMemcpy(xtemp, x, A->num_rows*sizeof(value_type), cudaMemcpyDeviceToDevice);

    // This code launches nLevels kernels on the GPU in stream 0. They are naturally synchronized.
    if (nnz_per_row <=  2)  { csr_tri_solve_cpu_level_scheduler_T<USECACHE,2>(*A, nLevels, maxNumUnknownsAtSameLevel, levelSum, levelsCPU, perms, x); }
    else if (nnz_per_row <=  4)  { csr_tri_solve_cpu_level_scheduler_T<USECACHE,4>(*A, nLevels, maxNumUnknownsAtSameLevel, levelSum, levelsCPU, perms, x); }
    else if (nnz_per_row <=  8)  { csr_tri_solve_cpu_level_scheduler_T<USECACHE,8>(*A, nLevels, maxNumUnknownsAtSameLevel, levelSum, levelsCPU, perms, x); }
    else if (nnz_per_row <=  16) { csr_tri_solve_cpu_level_scheduler_T<USECACHE,16>(*A, nLevels, maxNumUnknownsAtSameLevel, levelSum, levelsCPU, perms, x); }
    else if (nnz_per_row <=  32) { csr_tri_solve_cpu_level_scheduler_T<USECACHE,32>(*A, nLevels, maxNumUnknownsAtSameLevel, levelSum, levelsCPU, perms, x); }

    // This code launches a single kernel and uses atomic ops to synchronizes blocks. Need to be careful with this code.
    if (nnz_per_row <=  2)  { csr_tri_solve_gpu_level_scheduler_T<USECACHE,2>(*A, nLevels, maxNumUnknownsAtSameLevel, levelSum, levels, perms, xtemp); }
    else if (nnz_per_row <=  4)  { csr_tri_solve_gpu_level_scheduler_T<USECACHE,4>(*A, nLevels, maxNumUnknownsAtSameLevel, levelSum, levels, perms, xtemp); }
    else if (nnz_per_row <=  8)  { csr_tri_solve_gpu_level_scheduler_T<USECACHE,8>(*A, nLevels, maxNumUnknownsAtSameLevel, levelSum, levels, perms, xtemp); }
    else if (nnz_per_row <=  16) { csr_tri_solve_gpu_level_scheduler_T<USECACHE,16>(*A, nLevels, maxNumUnknownsAtSameLevel, levelSum, levels, perms, xtemp); }
    else if (nnz_per_row <=  32) { csr_tri_solve_gpu_level_scheduler_T<USECACHE,32>(*A, nLevels, maxNumUnknownsAtSameLevel, levelSum, levels, perms, xtemp); }

    std::vector<value_type> xcpu(A->num_rows);
    cudaMemcpy(&xcpu[0], x, A->num_rows*sizeof(value_type), cudaMemcpyDeviceToHost);
    std::vector<value_type> xcpu2(A->num_rows);
    cudaMemcpy(&xcpu2[0], xtemp, A->num_rows*sizeof(value_type), cudaMemcpyDeviceToHost);
    for (size_t i=0; i<A->num_rows; i++)
    std::cout << "err["<<i<<"]="<<fabs((xcpu[i]-xcpu2[i])/xcpu[i])<<"\t";
    std::cout << std::endl;
    std::cout << std::endl;
    cudaFree((void*) xtemp);
    return 0;

#endif

  return 1;
}


#endif // CSR_TRI_SOLVE_GPU_LEVEL_SCHEDULER_H
